#!/usr/bin/env scheme-script
;; Copyright © 2019 Gwen Weinholt        !#
;; SPDX-License-Identifier: MIT

;; Tests the GPT library. Test images are built by run-test.sh.

(import
  (rnrs)
  (fs partitions common)
  (fs partitions gpt)
  (uuid))

(define (make-sector-reader p sector-size)
  (lambda (lba)
    (set-port-position! p (* lba sector-size))
    (get-bytevector-n p sector-size)))

(define (part->list part)
  (list 'number (part-number part)
        'start (part-start-lba part)
        'end (part-end-lba part)
        'total (part-total-lba part)
        'type (part-type part)
        'guid (gptpart-guid part)
        'name (gptpart-name part)))

(display "Checking sexp->gpt with no parts\n")
(let ((gpt (sexp->gpt '((disk-id "739B17A4-1AAF-D74F-9EEE-77F40D34F15B")))))
  (validate-parttable gpt)
  (assert (equal? (parttable-disk-id gpt)
                  (string->uuid "739B17A4-1AAF-D74F-9EEE-77F40D34F15B")))
  (assert (null? (parttable-partitions gpt)))
  (assert (equal? 34 (gpt-first-usable-lba gpt)))
  (assert (equal? 34 (gpt-last-usable-lba gpt)))
  (assert (equal? (+ 34 32 1) (gpt-alternate-lba gpt))))
(newline)

(display "Checking sexp->gpt with 1 part\n")
(let ((gpt (sexp->gpt '((disk-id "739B17A4-1AAF-D74F-9EEE-77F40D34F15B")
                       (part (name "dummy")
                             (start 34) (end 40)
                             (id "B12A7328-581F-A1D2-0A4B-70A0C93EC931")
                             (type "C12A7328-F81F-11D2-BA4B-00A0C93EC93B"))))))
  (validate-parttable gpt)
  (assert (equal? (parttable-disk-id gpt)
                  (string->uuid "739B17A4-1AAF-D74F-9EEE-77F40D34F15B")))
  (assert (equal? 34 (gpt-first-usable-lba gpt)))
  (assert (equal? 40 (gpt-last-usable-lba gpt)))
  (assert (equal? (+ 40 32 1) (gpt-alternate-lba gpt)))
  (assert (equal? 1 (length (parttable-partitions gpt))))
  (let ((p (car (parttable-partitions gpt))))
    (assert (equal? "dummy" (gptpart-name p)))
    (assert (equal? 34 (part-start-lba p)))
    (assert (equal? 40 (part-end-lba p)))
    (assert (equal? (+ (- 40 34) 1) (part-total-lba p)))
    (assert (equal? (string->uuid "B12A7328-581F-A1D2-0A4B-70A0C93EC931")
                    (gptpart-guid p)))
    (assert (equal? (string->uuid "C12A7328-F81F-11D2-BA4B-00A0C93EC93B")
                    (part-type p)))))
(newline)

(display "Checking sexp->gpt with 2 parts\n")
(let ((gpt (sexp->gpt '((disk-id "739B17A4-1AAF-D74F-9EEE-77F40D34F15B")
                       (part (name "part1")
                             (start 34) (end 100)
                             (id "B12A7328-581F-A1D2-0A4B-70A0C93EC931")
                             (type "C12A7328-F81F-11D2-BA4B-00A0C93EC93B"))
                       (part (name "part2")
                             (start 101) (end 200)
                             (id "ABCD7328-581F-A1D2-0A4B-70A0C93EC931")
                             (type "EE2A7389-081F-D1D2-BA4B-00A0C93EC93B"))))))
  (validate-parttable gpt)
  (assert (equal? (parttable-disk-id gpt)
                  (string->uuid "739B17A4-1AAF-D74F-9EEE-77F40D34F15B")))
  (assert (equal? 34 (gpt-first-usable-lba gpt)))
  (assert (equal? 200 (gpt-last-usable-lba gpt)))
  (assert (equal? (+ 200 32 1) (gpt-alternate-lba gpt)))
  (assert (equal? 2 (length (parttable-partitions gpt))))
  (let ((p (car (parttable-partitions gpt))))
    (assert (equal? "part1" (gptpart-name p)))
    (assert (equal? 34 (part-start-lba p)))
    (assert (equal? 100 (part-end-lba p)))
    (assert (equal? (+ (- 100 34) 1) (part-total-lba p)))
    (assert (equal? (string->uuid "B12A7328-581F-A1D2-0A4B-70A0C93EC931")
                    (gptpart-guid p)))
    (assert (equal? (string->uuid "C12A7328-F81F-11D2-BA4B-00A0C93EC93B")
                    (part-type p))))
  (let ((p (cadr (parttable-partitions gpt))))
    (assert (equal? "part2" (gptpart-name p)))
    (assert (equal? 101 (part-start-lba p)))
    (assert (equal? 200 (part-end-lba p)))
    (assert (equal? (+ (- 200 101) 1) (part-total-lba p)))
    (assert (equal? (string->uuid "ABCD7328-581F-A1D2-0A4B-70A0C93EC931")
                    (gptpart-guid p)))
    (assert (equal? (string->uuid "EE2A7389-081F-D1D2-BA4B-00A0C93EC93B")
                    (part-type p)))))
(newline)

(display "Checking sexp->gpt with named part type\n")
(let ((gpt (sexp->gpt '((disk-id "739B17A4-1AAF-D74F-9EEE-77F40D34F15B")
                       (part (name "dummy")
                             (start 34) (end 40)
                             (type linux))))))
  (validate-parttable gpt)
  (assert (equal? (parttable-disk-id gpt)
                  (string->uuid "739B17A4-1AAF-D74F-9EEE-77F40D34F15B")))
  (assert (equal? 34 (gpt-first-usable-lba gpt)))
  (assert (equal? 40 (gpt-last-usable-lba gpt)))
  (assert (equal? (+ 40 32 1) (gpt-alternate-lba gpt)))
  (assert (equal? 1 (length (parttable-partitions gpt))))
  (let ((p (car (parttable-partitions gpt))))
    (assert (equal? "dummy" (gptpart-name p)))
    (assert (equal? 34 (part-start-lba p)))
    (assert (equal? 40 (part-end-lba p)))
    (assert (equal? (+ (- 40 34) 1) (part-total-lba p)))
    (assert (equal? (string->uuid "0FC63DAF-8483-4772-8E79-3D69D8477DE4")
                    (part-type p)))))
(newline)

(define (make-bytevector-writer disk)
  (lambda (data sector)
    (bytevector-copy! data 0 disk (fx* sector 512) (bytevector-length data))))

(define (make-bytevector-reader disk)
  (let ((tmp (make-bytevector 512 0)))
    (lambda (sector)
      (bytevector-copy! disk (fx* sector 512) tmp 0 512)
      tmp)))

(display "Checking create-gpt! with 1 part\n")
(let* ((disk (make-bytevector (fx* 512 100)))
       (gpt (sexp->gpt `((disk-id "739B17A4-1AAF-D74F-9EEE-77F40D34F15B")
                        (part (name "dummy")
                              (start 34) (end 40)
                              (id "B12A7328-581F-A1D2-0A4B-70A0C93EC931")
                              (type "C12A7328-F81F-11D2-BA4B-00A0C93EC93B"))))))
  (create-gpt! gpt (make-bytevector-reader disk) (make-bytevector-writer disk))

  (let-values (((primary backup) (read-gpt (make-bytevector-reader disk) #f)))
    (assert (equal? (parttable-disk-id primary)
                    (string->uuid "739B17A4-1AAF-D74F-9EEE-77F40D34F15B")))
    (assert (equal? (parttable-disk-id primary) (parttable-disk-id backup)))
    (assert (eqv? 1 (length (parttable-partitions primary))))
    (let ((p (car (parttable-partitions primary))))
      (assert (equal? "dummy" (gptpart-name p)))
      (assert (equal? 34 (part-start-lba p)))
      (assert (equal? 40 (part-end-lba p)))
      (assert (equal? (+ (- 40 34) 1) (part-total-lba p)))
      (assert (equal? (string->uuid "C12A7328-F81F-11D2-BA4B-00A0C93EC93B")
                      (part-type p))))))

(display "Checking create-gpt! with 2 parts\n")
(let* ((disk (make-bytevector (fx* 512 300)))
       (gpt (sexp->gpt `((part (name "some-very-long-partition-name-to-check-if-it-does-not-overlap")
                               (start 34) (end 40)
                               (type linux-swap))
                         (part (name "linux")
                               (id "B12A7328-581F-A1D2-0A4B-70A0C93EC931")
                               (start 41) (end 200)
                               (type linux))))))
  (create-gpt! gpt (make-bytevector-reader disk) (make-bytevector-writer disk))

  (let-values (((primary backup) (read-gpt (make-bytevector-reader disk) #f)))
    (assert (equal? (parttable-disk-id primary) (parttable-disk-id backup)))
    (assert (eqv? 2 (length (parttable-partitions primary))))
    (let ((p (car (parttable-partitions primary))))
      (assert (equal? "some-very-long-partition-name-to-che" (gptpart-name p)))
      (assert (equal? 34 (part-start-lba p)))
      (assert (equal? 40 (part-end-lba p)))
      (assert (equal? (+ (- 40 34) 1) (part-total-lba p)))
      (assert (equal? (string->uuid "0657FD6D-A4AB-43C4-84E5-0933C84B4F4F")
                      (part-type p))))
    (let ((p (cadr (parttable-partitions primary))))
      (assert (equal? "linux" (gptpart-name p)))
      (assert (equal? 41 (part-start-lba p)))
      (assert (equal? 200 (part-end-lba p)))
      (assert (equal? (+ (- 200 41) 1) (part-total-lba p)))
      (assert (equal? (string->uuid "B12A7328-581F-A1D2-0A4B-70A0C93EC931")
                      (gptpart-guid p)))
      (assert (equal? (string->uuid "0FC63DAF-8483-4772-8E79-3D69D8477DE4")
                      (part-type p))))))

(display "Checking testdisk1g.img\n")
(call-with-port (open-file-input-port "testdisk1g.img")
  (lambda (p)
    (let-values ([(primary backup) (read-gpt (make-sector-reader p 512) 2880)])
      (write primary)
      (newline)
      (write backup)
      (newline)
      (write (part->list (car (parttable-partitions primary))))
      (newline)
      (assert (equal? (parttable-disk-id primary)
                      (string->uuid "739B17A4-1AAF-D74F-9EEE-77F40D34F15B")))
      (assert (equal? (parttable-disk-id primary) (parttable-disk-id backup)))
      (assert (eqv? 1 (length (parttable-partitions primary))))
      (assert
       (equal?
        (part->list (car (parttable-partitions primary)))
        `(number 1 start 34 end 2847 total 2814
                 type ,(string->uuid "C12A7328-F81F-11D2-BA4B-00A0C93EC93B")
                 guid ,(string->uuid "C140638B-2F47-C548-8D9D-7A9EC8FE9BB2")
                 name "λoko")))
      (assert (equal? (map part->list (parttable-partitions primary))
                      (map part->list (parttable-partitions backup)))))))
(newline)

(display "Checking testdisk2g.img\n")
(call-with-port (open-file-input-port "testdisk2g.img")
  (lambda (p)
    (let-values ([(primary backup) (read-gpt (make-sector-reader p 512) 2880)])
      (write primary)
      (newline)
      (write backup)
      (newline)
      (write (part->list (car (parttable-partitions primary))))
      (newline)
      (assert (equal? (parttable-disk-id primary)
                      (string->uuid "00000001-0002-0003-0004-000000000005")))
      (assert (equal? (parttable-disk-id primary) (parttable-disk-id backup)))
      (assert
       (equal?
        (part->list (car (parttable-partitions primary)))
        `(number 127 start 34 end 2847 total 2814
                 type ,(string->uuid "aaaaaaab-bbbc-ddde-fff0-111111111112")
                 guid ,(string->uuid "22222223-3334-4445-5556-777777777778")
                 name "TheNameFieldCanBeAtMost36LettersLong")))
      (assert (equal? (map part->list (parttable-partitions primary))
                      (map part->list (parttable-partitions backup)))))))
(newline)

(display "Checking testdisk3g.img\n")
(call-with-port (open-file-input-port "testdisk3g.img")
  (lambda (p)
    (let-values ([(primary backup) (read-gpt (make-sector-reader p 4096) 2880)])
      (write primary)
      (newline)
      (write backup)
      (newline)
      (write (map part->list (parttable-partitions primary)))
      (newline)
      (assert (equal? (parttable-disk-id primary)
                      (string->uuid "C29A1A22-2DA6-4D4F-AA9B-1CD3B5990C7C")))
      (assert (equal? (parttable-disk-id primary) (parttable-disk-id backup)))
      (assert
       (equal?
        (map part->list (parttable-partitions primary))
        `((number 2 start 256 end 1000 total 745
                  type ,(string->uuid "C12A7328-F81F-11D2-BA4B-00A0C93EC93B")
                  guid ,(string->uuid "CDD33252-29F6-9F41-B87A-1C52FA1B5E4D")
                  name "")
          (number 128 start 1024 end 2875 total 1852
                  type ,(string->uuid "0FC63DAF-8483-4772-8E79-3D69D8477DE4")
                  guid ,(string->uuid "1667B437-1547-B345-A764-ED1609528F0E")
                  name ""))))
      (assert (equal? (map part->list (parttable-partitions primary))
                      (map part->list (parttable-partitions backup)))))))
(newline)

(flush-output-port (current-output-port))
